@@ -5,8 +5,7 @@ import matplotlib.pyplot as plt
 import matplotlib as mpl
 from matplotlib.colors import LinearSegmentedColormap
 
-from patsy.eval import EvalEnvironment
-
+import six
 import numpy as np
 import pandas as pd
 import warnings
@@ -14,10 +13,15 @@ import warnings
 from .aes import aes
 from .legend import make_legend
 from .themes import theme_gray
+from .themes import element_text
 from . import discretemappers
+from .utils import format_ticks, sorted_unique
+import urllib
+import base64
+import os
 
-# from PIL import Image
-
+if os.environ.get("GGPLOT_DEV"):
+    from PIL import Image
 
 class ggplot(object):
     """
@@ -35,7 +39,7 @@ class ggplot(object):
 
     Examples
     ----------
-    >>> p = ggplot(aes(x='x', y='y'), data=diamonds)
+    >>> p = ggplot(aes(x='carat', y='price'), data=diamonds)
     >>> print(p + geom_point())
     """
 
@@ -48,9 +52,11 @@ class ggplot(object):
             aesthetics, data = data, aesthetics
         self._aes = aesthetics
         self.data = data.copy()
-        self._evaluate_aes_expressions()
+        self._handle_index()
+        self.data = self._aes._evaluate_expressions(self.data)
         self.data = self._aes.handle_identity_values(self.data)
 
+
         self.layers = []
 
         # labels
@@ -67,22 +73,29 @@ class ggplot(object):
 
         # scales
         self.scales = []
-
         self.scale_identity  = set()
 
+        # margins
+        self.margins = None
+
         self.scale_x_log = None
         self.scale_y_log = None
 
         self.scale_x_reverse = None
         self.scale_y_reverse = None
 
+        # generic axis text
+        self.axis_text = None
+
         self.xbreaks = None
         self.xtick_labels = None
         self.xtick_formatter = None
+        self.x_axis_text = None
 
         self.ybreaks = None
         self.ytick_labels = None
         self.ytick_formatter = None
+        self.y_axis_text = None
 
         # faceting
         self.grid = None
@@ -102,44 +115,41 @@ class ggplot(object):
     def __repr__(self):
         self.make()
         # this is nice for dev but not the best for "real"
-        # self.fig.savefig('/tmp/ggplot.png', dpi=160)
-        # img = Image.open('/tmp/ggplot.png')
-        # img.show()
+        if os.environ.get("GGPLOT_DEV"):
+            self.fig.savefig('/tmp/ggplot-%d.png' % self.__hash__(), dpi=160)
+            img = Image.open('/tmp/ggplot-%d.png' % self.__hash__())
+            img.show()
+            return "<ggplot: (%d)>" % self.__hash__()
         plt.show()
         return "<ggplot: (%d)>" % self.__hash__()
 
-    def _evaluate_aes_expressions(self):
+    def show(self):
         """
-        Evaluates patsy expressions within the aesthetics. For example, 'x + 1'
-        , 'factor(x)', or 'pd.cut(price, bins=10)')
+        Builds and displays your plot.
         """
-        for key, item in self._aes.items():
-            if item not in self.data:
-                def factor(s, levels=None, labels=None):
-                    return s.apply(str)
+        self.make()
+        plt.show()
 
-                env = EvalEnvironment.capture(eval_env=(self._aes.__eval_env__ or 1)).with_outer_namespace({ "factor": factor, "pd": pd, "np": np })
-                try:
-                    new_val = env.eval(item, inner_namespace=self.data)
-                    self.data[item] = new_val
-                except:
-                    pass
+    def _handle_index(self):
+        if '__index__' in self._aes.values():
+            self.data['__index__'] = self.data.index
 
     def add_labels(self):
         labels = [(self.fig.suptitle, self.title)] #, (plt.xlabel, self.xlab), (plt.ylabel, self.ylab)]
         for mpl_func, label in labels:
             if label:
-                mpl_func(label)
+                if isinstance(label, (str, six.text_type)):
+                    label = element_text(label)
+                label.override(0.5, 0.95)
+                label.apply_to_fig(self.fig)
 
         if not self.facets:
             return
         if self.facets.is_wrap:
             return
         if self.facets.rowvar:
-            for row, name in enumerate(sorted(self.data[self.facets.rowvar].unique())):
-                if self.facets.is_wrap==True:
-                    continue
-                elif self.facets.colvar:
+            for row, name in enumerate(sorted_unique(self.data[self.facets.rowvar])):
+                if self.facets.colvar:
                     ax = self.subplots[row][-1]
                 else:
                     ax = self.subplots[row]
@@ -148,7 +158,7 @@ class ggplot(object):
                 ax.set_ylabel(name, fontsize=10, rotation=-90)
 
         if self.facets.colvar:
-            for col, name in enumerate(sorted(self.data[self.facets.colvar].unique())):
+            for col, name in enumerate(sorted_unique(self.data[self.facets.colvar])):
                 if len(self.subplots.shape) > 1:
                     col = col % self.facets.ncol
                     ax = self.subplots[0][col]
@@ -157,10 +167,11 @@ class ggplot(object):
                 ax.set_title(name, fontdict={'fontsize': 10})
 
     def apply_limits(self):
-        limits = [(plt.xlim, self.xlimits), (plt.ylim, self.ylimits)]
-        for mpl_func, limit in limits:
-            if limit:
-                mpl_func(limit)
+        for ax in self._iterate_subplots():
+            if self.xlimits:
+                ax.set_xlim(self.xlimits)
+            if self.ylimits:
+                ax.set_ylim(self.ylimits)
 
     def apply_scales(self):
         for scale in self.scales:
@@ -194,6 +205,9 @@ class ggplot(object):
                 self._aes.data['x'] = y
                 self._aes.data['y'] = x
 
+        if self.margins:
+            plt.subplots_adjust(**self.margins)
+
     def apply_axis_labels(self):
         if self.xlab:
             xlab = self.xlab
@@ -209,9 +223,10 @@ class ggplot(object):
                     ax.xaxis.set_ticklabels(self.xtick_labels)
         if self.xtick_formatter:
             for ax in self._iterate_subplots():
-                labels = [self.xtick_formatter(label) for label in ax.get_xticks()]
+                labels = []
+                for label_text in ax.get_xticks():
+                    labels.append(self.xtick_formatter(label_text))
                 ax.xaxis.set_ticklabels(labels)
-
         if self.ybreaks:
             for ax in self._iterate_subplots():
                 ax.yaxis.set_ticks(self.ybreaks)
@@ -219,20 +234,48 @@ class ggplot(object):
             if isinstance(self.ytick_labels, list):
                 for ax in self._iterate_subplots():
                     ax.yaxis.set_ticklabels(self.ytick_labels)
-
         if self.ytick_formatter:
             for ax in self._iterate_subplots():
                 labels = [self.ytick_formatter(label) for label in ax.get_yticks()]
                 ax.yaxis.set_ticklabels(labels)
 
-        self.fig.text(0.5, 0.05, xlab)
+        if self.axis_text:
+            for ax in self._iterate_subplots():
+                xticks = format_ticks(ax.get_xticks())
+                ax.set_xticklabels(xticks, **self.axis_text.kwargs)
+
+                yticks = format_ticks(ax.get_yticks())
+                ax.set_yticklabels(yticks, **self.axis_text.kwargs)
+
+        if self.x_axis_text:
+            for ax in self._iterate_subplots():
+                xticks = format_ticks(ax.get_xticks())
+                ax.set_xticklabels(xticks, **self.x_axis_text.kwargs)
+        if self.y_axis_text:
+            for ax in self._iterate_subplots():
+                yticks = format_ticks(ax.get_yticks())
+                ax.set_yticklabels(yticks, **self.y_axis_text.kwargs)
+
+        if xlab:
+            if isinstance(xlab, (str, six.text_type)):
+                xlab = element_text(xlab)
+
+            # enforce it to be an x-label
+            xlab.override(0.5, 0.05)
+            xlab.apply_to_fig(self.fig)
 
         if self.ylab:
             ylab = self.ylab
         else:
             ylab = self._aes.get('y', '')
 
-        self.fig.text(0.05, 0.5, ylab, rotation='vertical')
+        if isinstance(ylab, (str, six.text_type)):
+            ylab = element_text(ylab)
+
+        if ylab:
+            # enforce it to be a y-label
+            ylab.override(0.05, 0.5, dict(rotation='vertical'))
+            ylab.apply_to_fig(self.fig)
 
     def _iterate_subplots(self):
         """
@@ -297,6 +340,8 @@ class ggplot(object):
             mapping = discretemappers.linetype_gen()
         elif aes_type=="size":
             mapping = discretemappers.size_gen(self.data[colname].unique())
+        elif aes_type=="group":
+            mapping = discretemappers.identity_gen(self.data[colname].unique())
         return mapping
 
     def _construct_plot_data(self):
@@ -481,7 +526,7 @@ class ggplot(object):
         else:
             yield (self.subplots, group)
 
-    def save(self, filename, width=None, height=None):
+    def save(self, filename, width=None, height=None, dpi=180):
         """
         Save ggplot to a .png file.
 
@@ -492,7 +537,7 @@ class ggplot(object):
         width: int, float
             width of the plot in inches
         height: int, float
-            width of the plot in inches
+            height of the plot in inches
         """
         self.make()
         w, h = self.fig.get_size_inches()
@@ -503,6 +548,58 @@ class ggplot(object):
         self.fig.set_size_inches(w, h)
         self.fig.savefig(filename)
 
+    def save_as_base64(self, as_tag=False, width=None, height=None, dpi=180):
+        """
+        Save ggplot to a base64 encoded string.
+
+        Parameters
+        ----------
+        filename : string
+            filepath to save to
+        as_tag: bool
+            if true, returns an <img /> tag with the image uri as the src attribute
+        width: int, float
+            width of the plot in inches
+        height: int, float
+            height of the plot in inches
+        """
+        imgdata = six.StringIO()
+        self.save(imgdata, width=width, height=height, dpi=dpi)
+        imgdata.seek(0)  # rewind the data
+        uri = 'data:image/png;base64,' + urllib.quote(base64.b64encode(imgdata.buf))
+        if as_tag==True:
+            return '<img src = "%s"/>' % uri
+        else:
+            return uri
+
+    def _prep_layer_for_plotting(self, layer, facetgroup):
+        """
+        Some types of geoms (layer) need to be prepped before calling the plot
+        function on them. This function performs those perperations and then
+        returns a dictionary of **kwargs for the layer.plot function to use.
+        """
+        if layer.__class__.__name__=="geom_bar":
+            mask = True
+            df = layer.setup_data(self.data, self._aes, facets=self.facets)
+            if df is None:
+                return False
+            if self.facets:
+                facet_filter = facetgroup[self.facets.facet_cols].iloc[0].to_dict()
+                for k, v in facet_filter.items():
+                    mask = (mask) & (df[k]==v)
+                df = df[mask]
+
+            if 'fill' in self._aes:
+                fill_levels = self.data[self._aes['fill']].unique()
+            else:
+                fill_levels = None
+            return dict(x_levels=self.data[self._aes['x']].unique(), fill_levels=fill_levels, lookups=df)
+        elif layer.__class__.__name__ in ("geom_boxplot", "geom_violin", "geom_errorbar"):
+            x_levels = list(pd.Series(self.data[self._aes['x']].unique()).sort_values())
+            return dict(x_levels=x_levels)
+        else:
+            return dict()
+
     def make(self):
         "Constructs the plot using the methods. This is the 'main' for ggplot"
         plt.close()
@@ -524,25 +621,10 @@ class ggplot(object):
             for _, group in groups:
                 for ax, facetgroup in self.get_facet_groups(group):
                     for layer in self.layers:
-                        if layer.__class__.__name__=="geom_bar":
-                            mask = True
-                            df = layer.setup_data(self.data, self._aes, facets=self.facets)
-                            if self.facets:
-                                facet_filter = facetgroup[self.facets.facet_cols].iloc[0].to_dict()
-                                for k, v in facet_filter.items():
-                                    mask = (mask) & (df[k]==v)
-                                df = df[mask]
-
-                            if 'fill' in self._aes:
-                                fill_levels = self.data[self._aes['fill']].unique()
-                            else:
-                                fill_levels = None
-                            layer.plot(ax, facetgroup, self._aes, x_levels=self.data[self._aes['x']].unique(),
-                                fill_levels=fill_levels, lookups=df)
-                        elif layer.__class__.__name__ in ("geom_boxplot", "geom_violin"):
-                            layer.plot(ax, facetgroup, self._aes, x_levels=self.data[self._aes['x']].unique())
-                        else:
-                            layer.plot(ax, facetgroup, self._aes)
+                        kwargs = self._prep_layer_for_plotting(layer, facetgroup)
+                        if kwargs==False:
+                            continue
+                        layer.plot(ax, facetgroup, self._aes, **kwargs)
 
             self.apply_limits()
             self.add_labels()
